#!/usr/bin/perl
############################################################################
#
#  Alaska Satellite Facility DAAC
#  Sentinel Radiometric Terrain Correction using GAMMA
#
############################################################################
#
# Copyright (C) 2016 Alaska Satellite Facility
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
# 
############################################################################

use FileHandle;
use File::Basename;
use Getopt::Long;
use File::Copy;
use File::Glob ':glob';
use Cwd;
use Env qw(GAMMA_HOME);
use POSIX;

my $res = 10.0;
if (($#ARGV + 1) < 1) {die <<EOS ;}

usage: $0 <options> output
	output		Output RTC file name
	-e dem 		(option) specify a DEM file to use (with par file)
			e.g. big.dem to specify big.dem and big.dem.par
        -r res          (option) specify the output resolution (default $res meters)

EOS

my $dem = '';
GetOptions ('e=s' => \$dem,'r=f' => \$res);

my $output = $ARGV[0];
my $log = "$output.log";
my $log_dir = cwd();
$log = $log_dir."/".$log;

print "===================================================================\n";
print "                      Sentinel RTC Program\n";
print "===================================================================\n";
print "Processing resolution is $res meters\n";

# 
# Get the beam mode for this input file
#
my $mode = '';
my @list = glob "S1A_*.SAFE";
foreach my $file ( @list ) { 
  print "file is $file\n";
  $mode = substr($file,4,2); 
  $mode = lc($mode);
} 
if ($mode eq '') { die "Can't find a sentinel file to process\n"; }
else { print "Found beam mode $mode\n"; }

#
# Get the base resolution for this input file
#
my $annotation = `cat */*/s1a*.xml`;
my @ann = split /\n/, $annotation;   # split into lines
$base_res = 0.0;

foreach my $line (@ann) {
    if ($line =~ m/<rangePixelSpacing>(\S+)<\/rangePixelSpacing>/) {
 	$base_res = $1;
    }
}
if ($base_res == 0.0) { die "Can't find base resolution for this image\n"; }
else { print "Found base resolution of $base_res\n"; }

#
# Get the DEM file that matches this scene
# by calling OpenTopo to retrieve an SRTMGL1 DEM file
#
if ($dem eq '') {
  print "Getting a DEM file covering this SAR image\n";
  my @cc = get_cc();
  $min_lon = $cc[0];
  $max_lon = $cc[1];
  $min_lat = $cc[2];
  $max_lat = $cc[3];

  $dem = "big.dem";
  $parfile = "$dem.par";

  $cmd = "wget -Odem.tif \"http://opentopo.sdsc.edu/otr/getdem?demtype=SRTMGL1&west=$min_lon&south=$min_lat&east=$max_lon&north=$max_lat&outputFormat=GTiff\"";
  execute("$cmd","$log");

  #
  # Convert DEM File into UTM coordinates
  #
  $zone = get_zone($min_lon,$max_lon);
  $proj = "-t_srs '+proj=utm +zone=$zone +datum=WGS84'";
  execute("gdalwarp -r cubic $proj dem.tif tmpdem.tif","$log");

  #
  # Convert DEM file into GAMMA format
  #
  execute("utm2dem_i2.pl tmpdem.tif $dem $parfile","$log");

} else {
  print "Using DEM file $dem\n";
  $parfile = "$dem.par";
}

#
# Start processing the scene
#
my $num_vv = 0;
my @vv = glob "*/*/*vv*.tiff";
foreach my $v ( @vv ) { ++$num_vv; }
if ($num_vv == 1) { 
  print "\nFound VV polarization - processing\n\n"; 
  $pol = "vv";
  $main_pol = "vv";
  process_pol("$pol","$base_res","$res","$output","$mode","$log");
}

my $num_vh = 0;
my @vh = glob "*/*/*vh*.tiff";
foreach my $v ( @vh ) { ++$num_vh; }
if ($num_vh == 1) { 
  print "\nFound VH polarization - processing\n\n"; 
  $pol = "vh";
  process_2nd_pol("$pol","$base_res","$res","$output","$mode","$dem","$log");
}

my $num_hh = 0;
my @hh = glob "*/*/*hh*.tiff";
foreach my $h ( @hh ) { ++$num_hh; }
if ($num_hh == 1) { 
  print "\nFound HH polarization - processing\n\n"; 
  $pol = "hh";
  $main_pol = "hh";
  process_pol("$pol","$base_res","$res","$output","$mode","$log");
}

my $num_hv = 0;
my @hv = glob "*/*/*hv*.tiff";
foreach my $h ( @hv ) { ++$num_hv; }
if ($num_hv == 1) { 
  print "\nFound HV polarization - processing\n\n"; 
  $pol = "hv";
  process_2nd_pol("$pol","$base_res","$res","$output","$mode","$dem","$log");
}

print "\nDone!!!\n\n";
exit;

#
# Main routine to process a single pol of data
# Creates the main .tif file, the dem, ls_map, and inc_map
#
sub process_pol() {
  my ($pol,$base_res,$res,$output,$mode,$log) = @_;

  print "Ingesting GRD file into Gamma format\n";
  $cmd = "par_S1_GRD */*/*$pol*.tiff */*/*$pol*.xml */*/*/calibration-*$pol*.xml */*/*/noise-*$pol*.xml $output.$pol.GRD.par $output.$pol.GRD";
  execute($cmd,$log);

  #
  # Multilook the data if required 
  #
  $look_fact = int ($res/$base_res+0.5);
  if ($look_fact > 1) {
    print "Multi-looking GRD file\n";
    $cmd = "multi_look_MLI $output.$pol.GRD $output.$pol.GRD.par $output.$pol.mgrd $output.$pol.mgrd.par $look_fact $look_fact";
    execute($cmd,$log);
  } else {
    move("$output.$pol.GRD","$output.$pol.mgrd") or die ("ERROR $0: Move failed: $!");
    move("$output.$pol.GRD.par","$output.$pol.mgrd.par") or die ("ERROR $0: Move failed: $!");
  }   

  my $options = "-p -j -n 1 -q -c ";

  print "Running RTC process... initializing\n";
  $cmd = "mk_geo_radcal $output.$pol.mgrd $output.$pol.mgrd.par $dem $dem.par geo_$pol/area.dem geo_$pol/area.dem_par geo_$pol image $res 0 $options";
  execute($cmd,$log);
 
  print "Running RTC process... coarse matching\n";
  $cmd = "mk_geo_radcal $output.$pol.mgrd $output.$pol.mgrd.par $dem $dem.par geo_$pol/area.dem geo_$pol/area.dem_par geo_$pol image $res 1 $options";
  eval {
    execute($cmd,$log);
  } or do {
    print "$0: Determination of the initial offset failed, skipping initial offset\n";
  };

  print "Running RTC process... fine matching\n";
  $cmd = "mk_geo_radcal $output.$pol.mgrd $output.$pol.mgrd.par $dem $dem.par geo_$pol/area.dem geo_$pol/area.dem_par geo_$pol image $res 2 $options";

  eval {
    execute($cmd,$log);
  } or do {
    print "ERROR: Failed to determine offset model\n";
    exit -1;
  };

  print "Running RTC process... finalizing\n";
  $cmd = "mk_geo_radcal $output.$pol.mgrd $output.$pol.mgrd.par $dem $dem.par geo_$pol/area.dem geo_$pol/area.dem_par geo_$pol image $res 3 $options";
  execute($cmd,$log);

  #
  # Convert results into GEOTIFF files
  #
  chdir("geo_$pol");
  execute("data2geotiff area.dem_par image_0.ls_map 5 ${output}.ls_map.tif",$log);
  execute("data2geotiff area.dem_par image_0.inc_map 2 ${output}.inc_map.tif",$log);
  execute("data2geotiff area.dem_par area.dem 2 outdem.tif",$log);
  execute("gdal_translate -ot Int16 outdem.tif ${output}.dem.tif",$log);

  #
  # Move results to PRODUCT directory, renaming in the process
  #
  my $out = "../PRODUCT";
  mkdir($out);

  $outname = "s1a-$mode-rtc-$output-$pol"; 
  move("image_cal_map.mli.tif","$out/$outname.tif") or die "Move failed: image_cal_map.mli.tif -> $out/$outname.tif";

  $outname = "s1a-$mode-rtc-$output"; 
  move("${output}.ls_map.tif","$out/${outname}-ls_map.tif") or die "ERROR $0: Move failed: $!";
  move("${output}.inc_map.tif","$out/${outname}-inc_map.tif") or die "ERROR $0: Move failed: $!";
  move("${output}.dem.tif","$out/${outname}-dem.tif") or die "ERROR $0: Move failed: $!";

  chdir("..");
}

#
# terrain geocode the cross-pol channel using the refinement information from the main channel
# thus bypassing modes 0, 1, and 2 of mk_geo_radcal
# creates the main cross-pol tif file
#
sub process_2nd_pol() {
  my ($pol2,$base_res,$res,$output,$mode,$dem,$log) = @_;
  if ($pol2 eq "hv") { $pol1 = "hh"; }
  else { $pol1 = "vv"; }

  my $options = "-p -j -n 1 -q -c ";

  print "Cross-pol is $pol2; Main pol is $pol1\n";

  print "Ingesting GRD file into Gamma format\n";
  $cmd = "par_S1_GRD */*/*$pol2*.tiff */*/*$pol2*.xml */*/*/calibration-*$pol2*.xml */*/*/noise-*$pol2*.xml $output.$pol2.GRD.par $output.$pol2.GRD";
  execute($cmd,$log);

  #
  # Multi-look the data if necessary
  #
  $look_fact = int ($res/$base_res+0.5);
  if ($look_fact > 1) {
    print "Multi-looking GRD file\n";
    $cmd = "multi_look_MLI $output.$pol2.GRD $output.$pol2.GRD.par $output.$pol2.mgrd $output.$pol2.mgrd.par $look_fact $look_fact";
    execute($cmd,$log);
  } else {
    move("$output.$pol2.GRD","$output.$pol2.mgrd") or die ("ERROR $0: Move failed: $!");
    move("$output.$pol2.GRD.par","$output.$pol2.mgrd.par") or die ("ERROR $0: Move failed: $!");
  }   

  #
  # Setup file name space for call to mk_geo_radcal
  # By copying results from main pol, we save a lot of time
  #
  $geo_dir = "geo_$pol2";
  $gc0 = "$geo_dir/image_0.map_to_rdc";
  $ls_map = "$geo_dir/image_0.ls_map";
  $inc_map = "$geo_dir/image_0.inc_map";
  $sim_map = "$geo_dir/image_0.sim";
  $pix_map = "$geo_dir/image_0.pix_map";
  $diff_par = "$geo_dir/image.diff_par";

  $gc0_main = $gc0;
  $gc0_main =~ s/$pol2/$pol1/g;

  $ls_map_main = $ls_map;
  $ls_map_main =~ s/$pol2/$pol1/g;

  $pix_map_main = $pix_map;
  $pix_map_main =~ s/$pol2/$pol1/g;

  $sim_map_main = $sim_map;
  $sim_map_main =~ s/$pol2/$pol1/g;

  $inc_map_main = $inc_map;
  $inc_map_main =~ s/$pol2/$pol1/g;

  $diff_par_main = $diff_par;
  $diff_par_main =~ s/$pol2/$pol1/g;

  unless (-d $geo_dir){
    print "creating directory for output geocoded image products: $geo_dir\n";
    $exit = system("mkdir $geo_dir");
    $exit == 0 or die "ERROR $0: non-zero exit status: mkdir $geo_dir\n";
  }

  print "Copying diff par file $diff_par_main to $diff_par\n";
  copy("$diff_par_main","$diff_par") or die ("ERROR $0: Copy failed: $!");

  execute("ln -f $gc0_main $gc0",$log);
  execute("ln -f $ls_map_main $ls_map",$log);
  execute("ln -f $inc_map_main $inc_map",$log);
  execute("ln -f $sim_map_main $sim_map",$log);
  execute("ln -f $pix_map_main $pix_map",$log);

  #
  # Perform terrain correction on the cross-pol file
  #
  $cmd = "mk_geo_radcal $output.$pol2.mgrd $output.$pol2.mgrd.par $dem $dem.par geo_$pol1/area.dem geo_$pol1/area.dem_par $geo_dir image $res 3 $options";
  execute($cmd,$log);

  #
  # Save result into PRODUCT directory
  #
  chdir("$geo_dir");

  $outname = "s1a-$mode-rtc-$output-$pol2.tif";
  my $out = "../PRODUCT";

  print "Moving file image_cal_map.mli.tif to $out/$outname\n";
  move("image_cal_map.mli.tif","$out/$outname") or die "Move failed: image_cal_map.mli.tif -> ../$outname";

  chdir("..");

}


#
# Extract the corner coordinates from the annotation xml file
#
sub get_cc {

    my $annotation = `cat */*/s1a*.xml`;
    my @ann = split /\n/, $annotation;   # split into lines

    my $min_lat = 90;
    my $min_lon = 180;
    my $max_lat = -90;
    my $max_lon = -180; 

    foreach my $line (@ann) {
      if ($line =~ m/<latitude>(\S+)<\/latitude>/) {
        $max_lat = ($1 > $max_lat) ? $1 : $max_lat;
        $min_lat = ($1 < $min_lat) ? $1 : $min_lat;
      }
      if ($line =~ m/<longitude>(\S+)<\/longitude/) {
        $max_lon = ($1 > $max_lon) ? $1 : $max_lon;
        $min_lon = ($1 < $min_lon) ? $1 : $min_lon;
      }
    } 

    $dmax_lat = sprintf("%.8g",$max_lat);
    $dmin_lat = sprintf("%.8g",$min_lat);
    $dmax_lon = sprintf("%.8g",$max_lon);
    $dmin_lon = sprintf("%.8g",$min_lon);

    $dmax_lat = $dmax_lat + 0.15;
    $dmin_lat = $dmin_lat - 0.15;
    $dmax_lon = $dmax_lon + 0.15;
    $dmin_lon = $dmin_lon - 0.15;

    print "    found max latitude of $dmax_lat\n";
    print "    found min latitude of $dmin_lat\n";
    print "    found max longitude of $dmax_lon\n";
    print "    found min longitude of $dmin_lon\n";

    @parms = ("$dmin_lon","$dmax_lon","$dmin_lat","$dmax_lat");
    return @parms;
}

#
# Get the UTM zone number for this image
#
sub get_zone{
  my ($min_lon,$max_lon) = @_;
  my $center_lon = ($min_lon+$max_lon)/2;
  my $zf = ($center_lon+180)/6 + 1;
  my $zone = sprintf "%.0f", floor($zf);
  return $zone
}

  
#
# Execute a command, logging the results
#
sub execute{
  my ($command, $log) = @_;
  if (-e $log){open(LOG,">>$log") or die "\nERROR $0: cannot open log file: $log\n";}
  else {open(LOG,">$log") or die "\nERROR $0: cannot open log file: $log\n";}
  print "$command\n";
  print LOG ("\n${command}\n");
  close LOG;
  $exit = system("$command 1>> $log");
  $exit == 0 or die "ERROR $0: non-zero exit status: $command\n"
}
